//////////////////////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief Stores a mxn sparse matrix with p non-zero diagonals
/// \details This matrix makes no assumptions regarding symmetry or order of diagonals, but
///          these are important factors to consider when using the raw underlying values in
///          cuda kernels.
///
///          Example of how storage works for 3x3 matrix:
///
///               [1 2 3]    0 0[1 2 3]       [0 0 1 2 3]
///           A = [4 5 6] =>   0[4 5 6]0   => [0 4 5 6 0] + [-2 -1 0 1 2]
///               [7 8 9]       [7 8 9]0 0    [7 8 9 0 0]
///
///          I.e. the underlying data storage contains an mxp matrix where each column
///          contains the values of one of the diagonals, and the p-length vector contains
///          the offset of the diagonal into the original matrix, with 0 representing the main
///          diagonal, -ve for lower diagonals, and +ve for upper diagonals.
/// \author Frederik Lange
/// \date March 2018
/// \copyright Copyright (C) 2018 University of Oxford
//////////////////////////////////////////////////////////////////////////////////////////////

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#ifndef CBF_SPARSE_DIAGONAL_MATRIX_CUH
#define CBF_SPARSE_DIAGONAL_MATRIX_CUH

#include <vector>
#include <memory>
#include <string>

#include <thrust/device_vector.h>
#include <thrust/iterator/zip_iterator.h>

#include "miscmaths/bfmatrix.h"

namespace CBF
{
  class SparseDiagonalMatrix
  {
    public:
      // CONSTRUCTORS //
      /// Basic constructor
      SparseDiagonalMatrix (unsigned int n_rows, unsigned int n_cols,
                           const std::vector<int>& offsets);
      /// Copy constructor
      SparseDiagonalMatrix (const SparseDiagonalMatrix &obj);

      ///MEMBER FUNCTIONS //
      /// Get a pointer to the matrix values
      float *get_raw_pointer();
      /// Return vector of diagonal offsets
      std::vector<int> get_offsets() const;
      /// Save Matrix to file
      void save_matrix_to_text_file(std::string file_name);
      /// Convert to MISCMATHS::SparseBFMatrix csc format
      std::shared_ptr<MISCMATHS::BFMatrix> convert_to_sparse_bf_matrix(
          MISCMATHS::BFMatrixPrecisionType prec);
    protected:

    private:
      // DATAMEMBERS //
      unsigned int n_rows_;
      unsigned int n_cols_;
      /// The starting row for each diagonal
      std::vector<int> offsets_;
      /// The values of the matrix will be stored in a device_vector
      thrust::device_vector<float> d_values_;
  };
} ///namespace CBF
#endif // CBF_SPARSE_DIAGONAL_MATRIX_CUH
